from typing import TypedDict


class CompletionFnTelemetryContext(TypedDict):
    session_id: str
    conversation_id: str
    message_id: str
    simulation_name: str
    agent_name: str
    span_type: str


class RiskEvalTelemetryContext(TypedDict):
    session_id: str
    conversation_id: str
    message_id: str
    simulation_name: str
    agent_name: str
    span_type: str
    risk_name: str



import unittest
import os
import tempfile
from unittest import mock

os.environ["SNOWGLOBE_API_KEY"] = "test"

from src.snowglobe.client.src.config import Config  # noqa: E402

os.environ["SNOWGLOBE_API_KEY"] = ""


class TestConfig(unittest.TestCase):
    def test_get_api_key_env(self):
        with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
            config = Config()
            self.assertEqual(config.API_KEY, "testkey")

    def test_get_api_key_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("SNOWGLOBE_API_KEY=rcfilekey\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    config = Config()
                    self.assertEqual(config.API_KEY, "rcfilekey")

    def test_get_application_id_env(self):
        with mock.patch.dict(os.environ, {"SNOWGLOBE_APP_ID": "appid"}):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(config.APPLICATION_ID, "appid")

    def test_get_application_id_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("SNOWGLOBE_APP_ID=rcappid\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(config.APPLICATION_ID, "rcappid")

    def test_get_control_plane_url_env(self):
        with mock.patch.dict(os.environ, {"CONTROL_PLANE_URL": "http://env-url"}):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(config.CONTROL_PLANE_URL, "http://env-url")

    def test_get_control_plane_url_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("CONTROL_PLANE_URL=http://rc-url\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(config.CONTROL_PLANE_URL, "http://rc-url")

    def test_get_completions_per_interval_env(self):
        with mock.patch.dict(os.environ, {"COMPLETIONS_PER_SECOND": "42"}):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(config.CONCURRENT_COMPLETIONS_PER_INTERVAL, 42)

    def test_get_completions_per_interval_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("COMPLETIONS_PER_SECOND=99\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(config.CONCURRENT_COMPLETIONS_PER_INTERVAL, 99)

    def test_get_completions_interval_seconds_env(self):
        with mock.patch.dict(os.environ, {"COMPLETIONS_INTERVAL_SECONDS": "77"}):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(config.CONCURRENT_COMPLETIONS_INTERVAL_SECONDS, 77)

    def test_get_completions_interval_seconds_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("COMPLETIONS_INTERVAL_SECONDS=88\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(
                            config.CONCURRENT_COMPLETIONS_INTERVAL_SECONDS, 88
                        )

    def test_get_concurrent_risk_evaluations_env(self):
        with mock.patch.dict(os.environ, {"CONCURRENT_RISK_EVALUATIONS": "123"}):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(config.CONCURRENT_RISK_EVALUATIONS, 123)

    def test_get_concurrent_risk_evaluations_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("CONCURRENT_RISK_EVALUATIONS=321\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(config.CONCURRENT_RISK_EVALUATIONS, 321)

    def test_get_concurrent_risk_evaluations_interval_seconds_env(self):
        with mock.patch.dict(
            os.environ, {"CONCURRENT_RISK_EVALUATIONS_INTERVAL_SECONDS": "55"}
        ):
            with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                config = Config()
                self.assertEqual(
                    config.CONCURRENT_RISK_EVALUATIONS_INTERVAL_SECONDS, 55
                )

    def test_get_concurrent_risk_evaluations_interval_seconds_rcfile(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            rc_path = os.path.join(tmpdir, ".snowgloberc")
            with open(rc_path, "w") as f:
                f.write("CONCURRENT_RISK_EVALUATIONS_INTERVAL_SECONDS=66\n")
            with mock.patch.dict(os.environ, {}, clear=True):
                with mock.patch("os.getcwd", return_value=tmpdir):
                    with mock.patch.dict(os.environ, {"SNOWGLOBE_API_KEY": "testkey"}):
                        config = Config()
                        self.assertEqual(
                            config.CONCURRENT_RISK_EVALUATIONS_INTERVAL_SECONDS, 66
                        )


if __name__ == "__main__":
    unittest.main()



import asyncio
from typing import Awaitable, Callable, cast, Coroutine, Union
from .models import (
    CompletionFunctionOutputs,
    CompletionRequest,
    RiskEvaluationOutputs,
    RiskEvaluationRequest,
)
from .types import (
    CompletionFnTelemetryContext,
    RiskEvalTelemetryContext,
)


async def run_completion_fn(
    completion_fn: Union[
        Callable[[CompletionRequest], CompletionFunctionOutputs],
        Callable[[CompletionRequest], Awaitable[CompletionFunctionOutputs]],
    ],
    completion_request: CompletionRequest,
    telemetry_context: CompletionFnTelemetryContext,  # noqa
) -> CompletionFunctionOutputs:
    if asyncio.iscoroutinefunction(completion_fn):
        try:
            asyncio.get_running_loop()
            awaitable_completion_output = completion_fn(completion_request)
            completion_output = await awaitable_completion_output
        except RuntimeError:
            awaitable_completion_output: Coroutine[
                None, None, CompletionFunctionOutputs
            ] = completion_fn(completion_request)
            completion_output = asyncio.run(awaitable_completion_output)
    else:
        sync_completion_fn = cast(
            Callable[[CompletionRequest], CompletionFunctionOutputs], completion_fn
        )
        completion_output = sync_completion_fn(completion_request)
    return completion_output


async def run_risk_evaluation_fn(
    risk_evaluation_fn: Union[
        Callable[[RiskEvaluationRequest], RiskEvaluationOutputs],
        Callable[[RiskEvaluationRequest], Awaitable[RiskEvaluationOutputs]],
    ],
    risk_evaluation_request: RiskEvaluationRequest,
    telemetry_context: RiskEvalTelemetryContext,  # noqa
) -> RiskEvaluationOutputs:
    if asyncio.iscoroutinefunction(risk_evaluation_fn):
        try:
            asyncio.get_running_loop()
            awaitable_risk_eval = risk_evaluation_fn(risk_evaluation_request)
            risk_evaluation = await awaitable_risk_eval
        except RuntimeError:
            awaitable_risk_eval: Coroutine[None, None, RiskEvaluationOutputs] = (
                risk_evaluation_fn(risk_evaluation_request)
            )
            risk_evaluation = asyncio.run(awaitable_risk_eval)
    else:
        # pyright doesn't understand that the above check means the function is synchronous here...
        sync_risk_evaluation_fn = cast(
            Callable[[RiskEvaluationRequest], RiskEvaluationOutputs], risk_evaluation_fn
        )
        risk_evaluation = sync_risk_evaluation_fn(risk_evaluation_request)
    return risk_evaluation



import os
from importlib.metadata import version as importlib_version
from typing import Awaitable, Callable, Union
from functools import wraps

from .models import (
    CompletionRequest,
    CompletionFunctionOutputs,
    RiskEvaluationRequest,
    RiskEvaluationOutputs,
)
from .types import CompletionFnTelemetryContext, RiskEvalTelemetryContext

try:
    import mlflow  # type: ignore
    import mlflow.tracing  # type: ignore
    from databricks.sdk import WorkspaceClient

    mlflow.tracing.enable()
except ImportError:
    mlflow = None
    WorkspaceClient = None

SNOWGLOBE_VERSION = importlib_version("snowglobe")


def trace_completion_fn(
    *,
    session_id: str,
    conversation_id: str,
    message_id: str,
    simulation_name: str,
    agent_name: str,
    span_type: str,
):
    def trace_decorator(
        run_completion_fn: Callable[
            [
                Union[
                    Callable[[CompletionRequest], CompletionFunctionOutputs],
                    Callable[[CompletionRequest], Awaitable[CompletionFunctionOutputs]],
                ],
                CompletionRequest,
                CompletionFnTelemetryContext,
            ],
            Awaitable[CompletionFunctionOutputs],
        ],
    ) -> Callable[
        [
            Union[
                Callable[[CompletionRequest], CompletionFunctionOutputs],
                Callable[[CompletionRequest], Awaitable[CompletionFunctionOutputs]],
            ],
            CompletionRequest,
            CompletionFnTelemetryContext,
        ],
        Awaitable[CompletionFunctionOutputs],
    ]:
        disable_mlflow = os.getenv("SNOWGLOBE_DISABLE_MLFLOW_TRACING") or ""
        if mlflow is not None and disable_mlflow.lower() != "true":
            if not WorkspaceClient:
                raise RuntimeError(
                    "The databricks-sdk package must be installed for MLflow instrumentation!"
                )

            w = WorkspaceClient()
            current_user = w.current_user.me()

            formatted_sim_name = simulation_name.lower().replace(" ", "_")
            default_experiment_name = (
                f"/Users/{current_user.user_name}/{formatted_sim_name}"
            )

            mlflow_experiment_name = (
                os.getenv("MLFLOW_EXPERIMENT_NAME") or default_experiment_name
            )
            mlflow.set_experiment(mlflow_experiment_name)

            mlflow_active_model_id = os.getenv("MLFLOW_ACTIVE_MODEL_ID")
            if mlflow_active_model_id:
                mlflow.set_active_model(model_id=mlflow_active_model_id)
            else:
                mlflow.set_active_model(name=agent_name)

            span_attributes = {
                "snowglobe.version": SNOWGLOBE_VERSION,
                "type": span_type,
                "session_id": str(session_id),
                "conversation_id": str(conversation_id),
                "message_id": str(message_id),
                "simulation_name": simulation_name,
                "agent_name": agent_name,
            }

            @mlflow.trace(
                name=span_type,
                span_type=span_type,
                attributes=span_attributes,
            )
            @wraps(run_completion_fn)
            async def run_completion_fn_wrapper(
                completion_fn: Union[
                    Callable[[CompletionRequest], CompletionFunctionOutputs],
                    Callable[[CompletionRequest], Awaitable[CompletionFunctionOutputs]],
                ],
                completion_request: CompletionRequest,
                telemetry_context: CompletionFnTelemetryContext,
            ):
                try:
                    mlflow.update_current_trace(  # type: ignore
                        metadata={"mlflow.trace.session": str(session_id)},
                        tags={
                            "session_id": str(session_id),
                            "conversation_id": str(conversation_id),
                            "message_id": str(message_id),
                            "simulation_name": simulation_name,
                            "agent_name": agent_name,
                        },
                    )
                    response = await run_completion_fn(
                        completion_fn, completion_request, telemetry_context
                    )
                    return response
                except Exception as e:
                    raise e

            return run_completion_fn_wrapper
        else:
            return run_completion_fn

    return trace_decorator


def trace_risk_evaluation_fn(
    *,
    session_id: str,
    conversation_id: str,
    message_id: str,
    simulation_name: str,
    agent_name: str,
    span_type: str,
    risk_name,
):
    def trace_decorator(
        run_risk_evaluation_fn: Callable[
            [
                Union[
                    Callable[[RiskEvaluationRequest], RiskEvaluationOutputs],
                    Callable[[RiskEvaluationRequest], Awaitable[RiskEvaluationOutputs]],
                ],
                RiskEvaluationRequest,
                RiskEvalTelemetryContext,
            ],
            Awaitable[RiskEvaluationOutputs],
        ],
    ) -> Callable[
        [
            Union[
                Callable[[RiskEvaluationRequest], RiskEvaluationOutputs],
                Callable[[RiskEvaluationRequest], Awaitable[RiskEvaluationOutputs]],
            ],
            RiskEvaluationRequest,
            RiskEvalTelemetryContext,
        ],
        Awaitable[RiskEvaluationOutputs],
    ]:
        disable_mlflow = os.getenv("SNOWGLOBE_DISABLE_MLFLOW_TRACING") or ""
        if mlflow and disable_mlflow.lower() != "true":
            if not WorkspaceClient:
                raise RuntimeError(
                    "The databricks-sdk package must be installed for MLflow instrumentation!"
                )

            w = WorkspaceClient()
            current_user = w.current_user.me()

            formatted_sim_name = simulation_name.lower().replace(" ", "_")
            default_experiment_name = (
                f"/Users/{current_user.user_name}/{formatted_sim_name}"
            )

            mlflow_experiment_name = (
                os.getenv("MLFLOW_EXPERIMENT_NAME") or default_experiment_name
            )
            mlflow.set_experiment(mlflow_experiment_name)

            mlflow_active_model_id = os.getenv("MLFLOW_ACTIVE_MODEL_ID")
            if mlflow_active_model_id:
                mlflow.set_active_model(model_id=mlflow_active_model_id)
            else:
                mlflow.set_active_model(name=agent_name)

            span_attributes = {
                "snowglobe.version": SNOWGLOBE_VERSION,
                "type": span_type,
                "session_id": str(session_id),
                "conversation_id": str(conversation_id),
                "message_id": str(message_id),
                "simulation_name": simulation_name,
                "agent_name": agent_name,
                "risk_name": risk_name,
            }

            @mlflow.trace(
                name=span_type,
                span_type=span_type,
                attributes=span_attributes,
            )
            @wraps(run_risk_evaluation_fn)
            async def risk_evaluation_fn_wrapper(
                risk_evaluation_fn: Union[
                    Callable[[RiskEvaluationRequest], RiskEvaluationOutputs],
                    Callable[[RiskEvaluationRequest], Awaitable[RiskEvaluationOutputs]],
                ],
                risk_evaluation_request: RiskEvaluationRequest,
                telemetry_context: RiskEvalTelemetryContext,
            ):
                try:
                    mlflow.update_current_trace(  # type: ignore
                        metadata={"mlflow.trace.session": str(session_id)},
                        tags={
                            "session_id": str(session_id),
                            "conversation_id": str(conversation_id),
                            "message_id": str(message_id),
                            "simulation_name": simulation_name,
                            "agent_name": agent_name,
                            "risk_name": risk_name,
                        },
                    )
                    response = await run_risk_evaluation_fn(
                        risk_evaluation_fn, risk_evaluation_request, telemetry_context
                    )
                    return response
                except Exception as e:
                    raise e

            return risk_evaluation_fn_wrapper
        else:
            return run_risk_evaluation_fn

    return trace_decorator



from logging import getLogger

import httpx

from .config import config, get_api_key_or_raise
from .models import SnowglobeData, SnowglobeMessage

LOGGER = getLogger(__name__)


async def fetch_experiments(app_id: str = None) -> list[dict]:
    """
    Fetch experiments from the Snowglobe server.

    Returns:
        list[dict]: A list of experiments.
    """
    async with httpx.AsyncClient() as client:
        experiments_url = f"{config.CONTROL_PLANE_URL}/api/experiments"

        if app_id:
            experiments_url += f"?appId={app_id}"
        try:
            # get elapsed time for this request
            import time

            start_time = time.monotonic()
            experiments_response = await client.get(
                experiments_url,
                headers={"x-api-key": get_api_key_or_raise()},
                timeout=60.0,  # Set timeout to 60 seconds
            )
        except httpx.ConnectTimeout:
            elapsed_time = time.monotonic() - start_time
            raise Exception(
                f"Warning: Connection timed out while fetching experiments. Elapsed time: {elapsed_time:.2f} seconds. Polling will continue. If this persists please contact Snowglobe support."
            )

    if not experiments_response.status_code == 200:
        try:
            message = experiments_response.json().get("message")
        except Exception:
            message = experiments_response.text
        LOGGER.error(f"Error fetching experiments: {message}")
        raise Exception(
            f"{experiments_response.status_code} - {message or 'Unknown error'}"
        )
    experiments = experiments_response.json()
    return experiments


async def fetch_messages(*, test) -> list[SnowglobeMessage]:
    """
    Fetch messages from the Snowglobe server for a given test.

    Args:
        test (str): The test identifier.

    Returns:
        list[SnowglobeMessage]: A list of messages associated with the test.
    """
    # init messages
    messages = [
        SnowglobeMessage(
            role="user",
            content=test["prompt"],
            snowglobe_data=SnowglobeData(
                conversation_id=test["conversation_id"],
                test_id=test["id"],
            ),
        ),
    ]
    # if full turn append response
    if "response" in test and test["response"]:
        messages.append(
            SnowglobeMessage(
                role="assistant",
                content=test["response"],
                snowglobe_data=SnowglobeData(
                    conversation_id=test["conversation_id"],
                    test_id=test["id"],
                ),
            )
        )

    # build rest of messages
    parent_id = test.get("parent_test_id")
    async with httpx.AsyncClient() as client:
        while parent_id:
            # get parent test
            parent_test_response = await client.get(
                f"{config.CONTROL_PLANE_URL}/api/experiments/{test['experiment_id']}/tests/{parent_id}",
                headers={"x-api-key": get_api_key_or_raise()},
            )

            if not parent_test_response.status_code == 200:
                raise Exception(
                    f"Error fetching parent test {parent_id}: {parent_test_response.text}"
                )

            parent_test = parent_test_response.json()

            parent_id = parent_test.get("parent_test_id")

            messages.insert(
                0,
                SnowglobeMessage(
                    role="assistant",
                    content=parent_test["response"],
                    snowglobe_data=SnowglobeData(
                        conversation_id=parent_test["conversation_id"],
                        test_id=parent_test["id"],
                    ),
                ),
            )

            messages.insert(
                0,
                SnowglobeMessage(
                    role="user",
                    content=parent_test["prompt"],
                    snowglobe_data=SnowglobeData(
                        conversation_id=parent_test["conversation_id"],
                        test_id=parent_test["id"],
                    ),
                ),
            )

    return messages
